/**
 * 
 * Copyright (c) 2021, Catalina Sbert (main author), catalina.sbert@uib.es
 *                     Jose-Luis Lisani (minor changes), joseluis.lisani@uib.es
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */




#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "library/aux_lib/parser.h"
#include "library/aux_lib/libImageFormats.h"

#include "library/cs_lib/kernels.h"
#include "library/cs_lib/scalingfunctions.h"


using namespace std;


//Main function
// Implements Algorithm 2 in IPOL accompanying paper
//
//Global parameters:
//
//F: surround kernel type (refer to Table 1 in accompanying IPOL paper)
//   F_AG = Multi-Scale Retinex (MSR) kernel (linearized version of original MSR) ('AG'), 
//   F_IG = continuous average of Gaussians ('IG'), 
//   F_IE = continuous average of exponentials ('IE'), 
//   F_1  = continuous ACE kernel ('1'), 
//   F_2  = continuous Land kernel ('2')
//
//f: scaling function type (refer to Table 3 in accompanying IPOL paper)
//   f_L     = Linear ('L')
//   f_log   = Logarithm ('log')
//   f_alpha = Power ('alpha')
//   f_NR    = Naka-Rushton ('NR')
//   f_E     = Histogram based ('E')
//
//Surround kernel parameters (refer to Table 1 in accompanying IPOL paper)
//F_AG kernel parameters
//N: number of scales (N in Table 1)
//l: minimum scale (sigma_1 in Table 1)
//h: scale factor (S in Table 1)
//F_IG kernel parameters
//i: minimum scale (sigma_1 in Table 1)
//S: scale factor (S in Table 1)
//F_IE kernel parameters
//a: minimum scale factor (s in Table 1)
//b: maximum scale factor (S in Table 1)
//F_1 kernel parameters
//A: scale factor (s in Table 1)
//F_2 kernel parameters
//L: scale parameter (sigma in Table 1)
//
//Scaling function parameters (see Table 3 in accompanying IPOL paper)
//u, v: percentage of pixels that define the Min and Max parameters of the 
//      scaling function (see Algorithm 1 in accompanying IPOL paper)
//
//f_E parameters
//p: factor in exponent of f_E function (p in Table 3)
//
//f_alpha parameters
//e: exponent in f_alpha function (alpha in Table 3) 
//
//f_NR parameters
//w: constant factor in f_NR function (A in Table 3) 


int main(int argc,  char **argv)
{
    //Get input parameters
    
    std::vector <OptStruct *> options;
    OptStruct oF = {(char *) "F:", 0, (char *) "AG", NULL, (char *) "Kernel (AG, IG, IE, 1, 2)"};
    options.push_back(&oF);
    OptStruct of = {(char *) "f:", 0, (char *) "E", NULL, (char *) "Scaling function (L, log, alpha, NR, E)"};
    options.push_back(&of);
    
    //Parameters AG (Gauss) kernel
    OptStruct oN = {(char *) "N:", 0, (char *) "5", NULL, (char *) "F_AG kernel: number of scales"};
    options.push_back(&oN);
    OptStruct ol = {(char *) "l:", 0, (char *) "1", NULL, (char *) "F_AG kernel: minimum scale"};
    options.push_back(&ol);
	OptStruct oh  = {(char *) "h:", 1, (char *) "1", NULL, (char *) "F_AG kernel: factor of maximum scale (max scale=h*min(width, height))"};
    options.push_back(&oh);
    
    
    //Parameters F_1 (Ace)
    OptStruct oA  = {(char *) "A:", 0, (char *) "0.5", NULL, (char *) "F_1 kernel: scale parameter"};
    options.push_back(&oA);

    //Parameters F_2 (Land)
    OptStruct oL  = {(char *) "L:", 0, (char *) "0.5", NULL, (char *) "F_2 kernel: scale parameter"};
    options.push_back(&oL);
    
    //Parameters IG (IntGauss)
    OptStruct oi  = {(char *) "i:", 0, (char *) "0.5", NULL, (char *) "F_IG kernel: minimum scale"};
    options.push_back(&oi);
    OptStruct oS  = {(char *) "S:", 0, (char *) "1", NULL, (char *) "F_IG kernel: maximum scale =S*min(width, height)"};
    options.push_back(&oS);

    //Parameters IE (IntExp)
    OptStruct oa  = {(char *) "a:", 0, (char *) "0.5", NULL, (char *) "F_IE kernel: minimum scale =a/min(nx,ny) "};
    options.push_back(&oa);
    OptStruct ob  = {(char *) "b:", 0, (char *) "1", NULL, (char *) "F_IE kernel: maximum scale=b*min(nx,ny)  "};
    options.push_back(&ob);
    
    //Parameters Scaling Functions
    OptStruct ou  = {(char *) "u:", 0, (char *) "1", NULL, (char *) "Scaling Function: percentage of pixels clipped to 0"};
    options.push_back(&ou);
    OptStruct ov  = {(char *) "v:", 0, (char *) "1", NULL, (char *) "Scaling Function: percentage of pixels clipped to 255"};
    options.push_back(&ov);
    
    //Parameters Histogram based scaling function
    OptStruct op  = {(char *) "t:", 0, (char *) "2", NULL, (char *) "Histogram based Scaling Function (f_E): power parameter"};
    options.push_back(&op);

    //Parameters Power scaling function
    OptStruct oe  = {(char *) "e:", 0, (char *) "0", NULL, (char *) "Power Scaling Function (f_alpha): power parameter (if 0, then estimate automatically)"};
    options.push_back(&oe);
    
    //Parameters Naka-Rushton scaling function
    OptStruct ow  = {(char *) "w:", 0, (char *) "0", NULL, (char *) "Naka-Rushton Scaling Function (f_NR): constant factor (if 0, then estimate automatically)"};
    options.push_back(&ow);

        
    //Input and output images
    std::vector<ParStruct *> pparameters;
    ParStruct pinput = {(char *) "input", NULL, (char *) "input file (TIFF or PNG format)"};
    pparameters.push_back(&pinput);  
    ParStruct poutput = {(char *) "output", NULL, (char *) "output file (TIFF or PNG format)"};
    pparameters.push_back(&poutput);  
      
    if (!parsecmdline((char *) "cs_filter", (char *) "Center/Surround filtering",
                      argc, argv, options, pparameters))
        return EXIT_FAILURE;
        
    
    char *kerneltype = oF.value;
    char *sfunc = of.value;
    
    //Check valid values for surround kernel and scaling function
    if (strcmp(kerneltype, "AG") && 
        strcmp(kerneltype, "IG") &&
        strcmp(kerneltype, "IE") &&
        strcmp(kerneltype, "1") &&
        strcmp(kerneltype, "2")) {
        printf("Error! Invalid kernel type. Available options: AG, IG, IE, 1, 2\n");
        return EXIT_FAILURE;
    }

    if (strcmp(sfunc, "L") && 
        strcmp(sfunc, "log") &&
        strcmp(sfunc, "alpha") &&
        strcmp(sfunc, "NR") &&
        strcmp(sfunc, "E")) {
        printf("Error! Invalid scaling function type. Available options: L, log, alpha, NR, E\n");
        return EXIT_FAILURE;
    }

    char *namein = pinput.value;
    char *nameout = poutput.value;
    
    printf("Input=%s   kernel=%s  scaling function=%s\n", namein, kerneltype, sfunc);
    
    //Read input data
    float *data_in;
    int width, height, nchannels;
    data_in=iio_read_image_float_split(namein, &width, &height, &nchannels);
    
    //do not use alpha channel
    if(nchannels >= 3) nchannels=3; 
    else nchannels=1;

    int minsize = (width < height)?(width):(height);
    int size = width*height;
    
    //Apply Center/Surround
    
    double *input= new double[nchannels*size];
    //convert input float data to double format
    for (int n=0; n < size*nchannels; n++) input[n] = (double) data_in[n];

    //Apply surround kernel: 
    
    double *outputCS; // outputCS = I/I*F
    
    if (!strcmp(kerneltype, "AG")) { //apply F_AG
        //Parameters F_AG kernel      
        int F_AG_N = atoi(oN.value);
        double F_AG_scale0 = (double) atof(ol.value);        
        double F_AG_sizefactor=(double) atof(oh.value);
      
        printf("F_AG kernel parameters: gauss_N=%i, gauss_l=%2.2f, gauss_f=%2.2f\n", 
                atoi(oN.value), atof(ol.value), atof(oh.value));    

        //compute scales
        double F_AG_scales[F_AG_N];
        F_AG_scales[0] = F_AG_scale0;
        double F_AG_w=1./F_AG_N;        
	    double F_AG_sizef=F_AG_sizefactor*((double) minsize);
	    double F_AG_factor=pow(F_AG_sizef / F_AG_scales[0], 1.0/((double) F_AG_N-1));
        for(int n=1; n< F_AG_N; n++) F_AG_scales[n]=F_AG_scales[n-1]*F_AG_factor;
                                                       
        outputCS = apply_kernel_AG(input, F_AG_scales, F_AG_N, F_AG_w, 
                                   width, height, nchannels); 
    } else {
    
        double *kernel=NULL; 
    
        if (!strcmp(kerneltype, "IG")) {  //apply F_IG
            //Parameters F_IG kernel   
            double F_IG_si, F_IG_S;
            F_IG_si = (double) atof(oi.value);
            F_IG_S = (double) atof(oS.value);
            printf("F_IG kernel parameter: F_IG_si=%2.2f, F_IG_S=%2.2f\n", 
                    atof(oi.value), atof(oS.value));    
                               
	        double sf = F_IG_S*(double) minsize;

            //compute kernel values
            kernel = F_IG(F_IG_si, sf, width, height);        
        }    
        
        
        if (!strcmp(kerneltype, "IE")) {  //apply F_IE
            //Parameters F_IE kernel   
            double F_IE_s, F_IE_S;
            F_IE_s = (double) atof(oa.value);
            F_IE_S = (double) atof(ob.value);
            printf("F_IE kernel parameter: F_IE_s=%2.2f, F_IE_S=%2.2f\n", 
                    atof(oa.value), atof(ob.value));    
                    
	        double si = F_IE_s/(double) minsize;
	        double sf = F_IE_S*(double) minsize;
	                           
            //compute kernel values
            kernel = F_IE(si, sf,  width, height);        
        }    
        
        if (!strcmp(kerneltype, "1")) {  //apply F_1
            //Parameters Ace Kernel
            double F_1_s = (double) atof(oA.value);
            printf("F_1 kernel parameter: F_1_s=%2.2f\n", atof(oA.value));    
        
            F_1_s = F_1_s/(double) minsize;
            
            //compute kernel values
            kernel = F_1_2(1, F_1_s, width, height);            
        }

        if (!strcmp(kerneltype, "2")) {  //apply F_2
            //Parameters Land Kernel
            double F_2_s = (double) atof(oL.value);
            printf("F_2 kernel parameter: F_2_s=%2.2f\n", atof(oL.value));    
                
            //compute kernel values
            kernel = F_1_2(2, F_2_s, width, height);            
        }
        
        
        outputCS = apply_kernel(input, kernel, width, height, nchannels);
        
        delete[] kernel;
    
    }
    
    //Apply scaling function: map [Min, Max] range on input to [0, 255]
    //                        on a channel-by-channel basis
    
    double *output = new double[nchannels*size]; //output=f(outputCS)
    
    //Compute Min and Max 
    float s1 = atof(ou.value);
    float s2 = atof(ov.value);

    printf("Parameters scaling function Min, Max: s1=%2.2f, s2=%2.2f\n", s1, s2);

    double Min, Max;
    if (nchannels == 3) 
        compute_min_max(Min, Max, outputCS, outputCS+size, outputCS+2*size, 
                        size, s1, s2); 
    else
        compute_min_max(Min, Max, outputCS, size, s1, s2); 
        
    printf("Min=%2.2f   Max=%2.2f\n", Min, Max);
    
    //Apply function
        
    if (!strcmp(sfunc, "L")) { //apply f_L (Linear)
        printf("Is f_L scaling function\n");
        for (int n=0; n < nchannels; n++)
            f_L(output+n*size, outputCS+n*size, size, Min, Max); 
    }
    
    if (!strcmp(sfunc, "log")) {  //apply f_log (Log)
        printf("Is f_log scaling function\n");
        for (int n=0; n < nchannels; n++)
            f_log(output+n*size, outputCS+n*size, size, Min, Max); 	
    }
    
    if (!strcmp(sfunc, "alpha")) {  //apply f_alpha (Power)
        printf("Is f_alpha scaling function\n");
        float f_alpha_e = atof(oe.value);
        printf("f_alpha function parameters: f_alpha_e=%2.2f\n", f_alpha_e);                

        if (f_alpha_e == 0) //estimate parameter automatically 
            f_alpha_e = estimate_parameter_alpha(outputCS, size, nchannels, 
                                                 Min, Max); 

        for (int n=0; n < nchannels; n++)
            f_alpha(output+n*size, outputCS+n*size, f_alpha_e, size, Min, Max); 	          
    }
    
    if (!strcmp(sfunc, "NR")) {  //apply f_NR (Naka-Rushton)
        printf("Is f_NR scaling function\n");
        float f_NR_A = atof(ow.value);
        printf("f_NR parameters: f_NR_A=%2.2f\n", f_NR_A);       
        
        if (f_NR_A == 0) //estimate parameter automatically 
            f_NR_A = estimate_parameter_NR(outputCS, size, nchannels, Min, Max); 
                 
        for (int n=0; n < nchannels; n++)
            f_NR(output+n*size, outputCS+n*size, f_NR_A, size, Min, Max); 
    }
    
    if (!strcmp(sfunc, "E")) { //apply f_E (Histogram-based)
        printf("Is f_E scaling function\n");
        float f_E_p = atof(op.value);
        printf("f_E function parameters: mai_p=%2.2f\n", f_E_p);        
        for (int n=0; n < nchannels; n++)
            f_E(output+n*size, outputCS+n*size, f_E_p, size, Min, Max); 
    }
    
    //output file
    float *data_out=new float[nchannels*size];
    for (int n=0; n < size*nchannels; n++) data_out[n] = (float) output[n];
    iio_save_image_float_split(nameout, data_out, width, height, nchannels);
    
    
    //clean up
    delete[] input;
    delete[] outputCS;
    delete[] output;
    delete[] data_out;
    free(data_in);
    
    return EXIT_SUCCESS;
    
}




