/**
 * 
 * Copyright (c) 2021, Catalina Sbert (main author), catalina.sbert@uib.es
 *                     Jose-Luis Lisani (minor changes), joseluis.lisani@uib.es
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fftw3.h>

// We supose that the input image (N X M array) is extended symmetrically 
// across its sides to a  2N X 2M array. Then we obtain a function wich is 
// even around (-0.5 , j) ,(N-0.5,j ), (i,-0.5) and (i, M-0.5). 
 
//Forward Fast Fourier Transform: out=FFT(in)
double *FTdata(double *out, double *in, int nx, int ny)
{
    fftw_plan p;
    
    // compute the cosinus transform of the input array
    p= fftw_plan_r2r_2d(ny, nx, in, out ,FFTW_REDFT10,FFTW_REDFT10, FFTW_ESTIMATE);
	fftw_execute(p);
    
    fftw_destroy_plan(p);

    return out;
}

//Inverse Fast Fourier Transform: out=IFFT(in)
double *IFTdata(double *out, double *in, int nx, int ny)
{    
    int l, dim4;
    fftw_plan p;
    
    dim4=4*nx*ny;
    
    // compute the inverse cosinus transform of the input array
    p= fftw_plan_r2r_2d(ny, nx, in, out ,FFTW_REDFT01,FFTW_REDFT01, FFTW_ESTIMATE);
	fftw_execute(p);
    
    for(l=0; l< nx*ny;l++) out[l]/=dim4;
    
    fftw_destroy_plan(p);
        
    return out;
}

//Compute convolution as
// out = data * kernel = IFFT(FFT(data) . FFT(kernel))
void convolution(double *out, double *data, double *kernel, int nx, int ny)
{
        
    int size = nx * ny;
    
    double *fft_data = (double*) fftw_malloc(sizeof(double) * size);
    double *fft_kernel = (double*) fftw_malloc(sizeof(double) * size);
    double *fft_out = (double*) fftw_malloc(sizeof(double) * size);
    
    FTdata(fft_data, data, nx, ny);
    
    FTdata(fft_kernel, kernel, nx, ny);
    
    for(int l=0; l <size; l++) fft_out[l]=fft_data[l]*fft_kernel[l];
    
    IFTdata(out, fft_out, nx, ny);
	
    fftw_free(fft_data);
    fftw_free(fft_kernel);
    fftw_free(fft_out);
   
}

    
//Compute convolution with Gaussian kernel at a given scale: 
//out = data * GaussianKernel(scale) = IFFT(FFT(data) . FFT(GaussianKernel))
void convolution_Gauss(double *data, double *out, 
                          double scale, int nx, int ny)
{
    int size = nx * ny;
    
    double *fft_data = (double*) fftw_malloc(sizeof(double) * size);
    
    /*compute the Fourier transform of the input data*/
    FTdata(fft_data, data, nx, ny);
    
    /*define the gaussian constants for the convolution*/
    
    double sigma=scale*scale/2.;
    double normx=M_PI/(double)nx;
    double normy=M_PI/(double) ny;
    normx*=normx;
    normy*=normy;
        
    int index;
    for(int j=0; j < ny; j++) {
        index=j*nx;
        for(int i=0; i < nx; i++)
            fft_data[i+index]*=exp((double)(-sigma)*(normx*i*i+normy*j*j));
    }
    
    /*compute the Inverse Fourier transform */
    IFTdata(out, fft_data, nx, ny);
        
    fftw_free(fft_data);
    
}



    

