/**
 * 
 * Copyright (c) 2021, Catalina Sbert (main author), catalina.sbert@uib.es
 *                     Jose-Luis Lisani (minor changes), joseluis.lisani@uib.es
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <fftw3.h>

#include "fftconvolutions.h"

#define EPS 0.00000001

//compute I/(I*F) 
//where I is an input image and F is a kernel 
//whose values are defined by the 'kernel' array
double *apply_kernel(double *input, double *kernel,
                     int width, int height, int nchannels)
{
    int size = width*height;

    //allocate memory
    double *output = new double[size*nchannels];
    
    //auxiliary variable
    double *dataconv = new double[size*nchannels];
    
    //perform operation on a channel-by-channel basis
    for (int n=0; n< nchannels; n++) {
        //compute I*F
        convolution(dataconv+n*size, input+n*size, kernel, width, height);
        //compute I/(I*F) 
        for (int l=0; l < size; l++) 			
		    output[l+n*size] = input[l+n*size]/(dataconv[l+n*size]+EPS);			    
    }
    
    delete[] dataconv;
    
    return output;
}


//compute I/(I*F) 
//where I is an input image and F = F_AG
//F_AG is the Multi-Scale Retinex (MSR) kernel
//(see Table 1 in IPOL accompanying paper)
double *apply_kernel_AG(double *input, double *scale, int nscales, double w,
                        int width, int height, int nchannels)
{
    int size = width*height;
    
    //allocate memory
    double *output = new double[size*nchannels];
    
    //auxiliary variables
    double *gaussconv = new double[size*nchannels];
    double *acumconv = new double[size*nchannels];
    
    for (int n=0; n< nchannels; n++) {   
        //compute I*F     
        memset(acumconv+n*size, 0, size*sizeof(double));
        //apply convolution at each scale and add (weighted) results
        for(int m=0; m < nscales; m++) {
            //compute convolution with Gaussian at scale scale[m]
            convolution_Gauss(input+n*size, gaussconv, scale[m], width, height);
            //accumulate convolutions
            for (int l=0; l < size; l++)
                acumconv[l+n*size] += w*gaussconv[l];  
        }

        //compute I/(I*F) 
        for (int l=0; l < size; l++) 			
		    output[l+n*size] = input[l+n*size]/(acumconv[l+n*size]+EPS);				    		        
    }
    
    delete[] gaussconv;
    delete[] acumconv;
    
    return output;
}

//compute values for kernel F_IG 
//(see Table 1 in IPOL accompanying paper)
double *F_IG(double sigma1, double sigma2, int width, int  height)
{
    int size = width*height;    
    //allocate memory
    double *kernel = new double[size];

    int i, j, l;
    
    double r, norm0;
    
    double sum;
   
    sigma1*=sigma1;
    
    sigma2*=sigma2;
    
    norm0=0.5*(1/sigma1-1/sigma2);
    

    for (j=0; j < height; j++){
        l=j*width;
        for (i=0; i < width; i++){
            if (i==0 && j==0) kernel[0]=norm0;
            else{
                r = (double) (i*i+j*j);
                kernel[l+i] = (exp(-r/(2.*sigma2))-exp(-r/(2.*sigma1)))/r;
            }
            
        }
    }

    sum=0.;
    for(l=0; l< width*height;l++)
        sum += kernel[l];
    
    for(l=0; l< width*height; l++) kernel[l] /= (4*sum); 
       
    return kernel;
}

    
//compute values for kernel F_IE 
//(see Table 1 in IPOL accompanying paper)
double *F_IE(double sigma1, double sigma2, int width, int height)
{
    int size = width*height;    
    //allocate memory
    double *kernel = new double[size];

    int i, j, l;
    double  norm0, r;
    double sum;
    
    norm0=1/sigma1-1/sigma2;
    
    for(j=0; j < height;j++){
        l=j*width;
        for(i=0;i< width; i++){
            if (i==0 && j==0) kernel[0]=norm0; 
            else{
                r = sqrt((double)(i*i+j*j));
                kernel[l+i] = (exp(-r/sigma2)-exp(-r/sigma1))/r;
            }
        }
    }
    
    sum=0.;
    for(l=0; l < width*height;l++)
        sum += kernel[l];
    
    
    for(l=0; l < width*height; l++) kernel[l] /= (4*sum); 
    
    return kernel;
    
}


//compute values for kernel F_1 and F_2 
//(see Table 1 in IPOL accompanying paper)
//if p = 1, then compute F_1
//if p = 2, then compute F_2
double *F_1_2(double p, double sigma, int width, int height)
{
    int size = width*height;    
    //allocate memory
    double *kernel = new double[size];

    int i, j, l;
    double r;
    double sum;
        
    for (j=0; j < height; j++){
        l=j*width;
        for (i=0; i < width; i++){
            r = pow(sqrt((double)(i*i+j*j))/sigma, p);
            kernel[i+l] = 1./(1+r);
        }
    }
    
    sum=0.;
    for (l=0; l < width*height; l++)
        sum += kernel[l];
    
    for(l=0; l< width*height; l++) kernel[l] /= (4*sum);  
    
    
    return kernel;
}


