/**
 * 
 * Copyright (c) 2021, Catalina Sbert (main author), catalina.sbert@uib.es
 *                     Jose-Luis Lisani (minor changes), joseluis.lisani@uib.es
 * All rights reserved.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define MAX(a,b) ((a) > (b) ? (a): (b))

#define MAX3(a,b,c) ((a > b) ? ((a > c)?(a):(c)):((b > c)?(b):(c)))
#define MIN3(a,b,c) ((a < b) ? ((a < c)?(a):(c)):((b < c)?(b):(c)))



//auxiliary function needed to sort double values in increasing order
int myComparisonFunction(const void *x, const void *y) {
    
    /* x and y are pointers to doubles.
     Returns  -1 if x < y
     0 if x == y
     +1 if x > y*/
    
    double dx, dy;
    
    dx = *(double *)x;
    dy = *(double *)y;
    
    if (dx < dy) {
        return -1;
    } else if (dx > dy) {
        return +1;
    }
    return 0;
}


//Implements Algorithm 1 for color images (RGB values)
//Computes Min and Max values that define the range of values [Min, Max]
//that the scaling function maps to [0, 255]
void compute_min_max(double &min, double &max, 
                     double *dataR, double *dataG, double *dataB, 
                     int size, float s1, float s2)
{

    double *datamin = new double[size];
    double *datamax = new double[size];
    for (int n=0; n < size; n++) {
        datamin[n]=MIN3(dataR[n], dataG[n], dataB[n]);
        datamax[n]=MAX3(dataR[n], dataG[n], dataB[n]);
    }
    
    qsort(datamin, size, sizeof(double), &myComparisonFunction);
    qsort(datamax, size, sizeof(double), &myComparisonFunction);
    
    int per1 = (int) (s1 * (float) size/100.0f);
    int per2 = (int) (s2 * (float) size/100.0f);
    
    min=datamin[per1];   
    max=datamax[size-1-per2];
    
    delete[] datamin;
    delete[] datamax;
}

//Implements Algorithm 1 for graylevel images (intensity values)
//Computes Min and Max values that define the range of values [Min, Max]
//that the scaling function maps to [0, 255]
void compute_min_max(double &min, double &max, 
                     double *data, 
                     int size, float s1, float s2)
{

    double *datasort = new double[size];
    memcpy(datasort, data, size*sizeof(double));
    
    qsort(datasort, size, sizeof(double), &myComparisonFunction);
    
    int per1 = (int) (s1 * (float) size/100.0f);
    int per2 = (int) (s2 * (float) size/100.0f);
    
    min=datasort[per1];
    max=datasort[size-1-per2];
    
    delete[] datasort;
}




//Auxiliary function needed to compute the Median of a set of values 
double Median(double *u, int size)
{
    
    if (size == 1) return u[0];
    
    double med;
    
    double *sort_u=new double[size];
    
    memcpy(sort_u, u, size*sizeof(double));
    
    qsort(sort_u, size, sizeof sort_u[0], &myComparisonFunction);
    
    int dsize = (size - 1) / 2;
    if (size % 2 == 1)  med=sort_u[dsize];
    else med= 0.5 * (sort_u[dsize]+sort_u[dsize+1]);
    
    delete[] sort_u;
    
    return med;
}



//Discrete implementation of f_E (Histogram based) scaling function 
//(see Table 3 in IPOL accompanying paper)
//Parameters:
//p: factor in exponent of Equation 12 (see IPOL accompanying paper)
void f_E(double *output, double *input, double p, 
                 int size, double Min, double Max)
{
    double t=p+1;

    //Hard-coded parameters for the discretization of the 
    //histogram of input values
    int nb=2;
    int npoints=256;   

    int nbins=nb*npoints;
        
    double *H=new double[nbins];
    double *S=new double[npoints];
    
    double *L=new double[npoints+1];
    double *PH=new double[npoints];
    
    double *sortdata=new double[size];
        
    double sum=0.;
    
    //define histogram bins
    double dh= (Max-Min)/nbins;
    double step=(Max-Min)/npoints;
    for(int k=0; k<= npoints; k++) L[k]=Min+k*step;


    memcpy(sortdata, input, size*sizeof(double));   
    qsort(sortdata, size, sizeof sortdata[0], &myComparisonFunction);
      
    //compute histogram
    int l=0;
    for(int b=0; b< nbins; b++){
        H[b]=0.;
        while(sortdata[l] < Min && l< size) l++;
        while(Min+b*dh <= sortdata[l] && sortdata[l]< Min+(b+1)*dh && l< size){
            H[b]++;
            l++;
        }
    }
    
    //normalize histogram
    for(int i=0; i< nbins; i++)
        H[i]/=size;
    
    //compute cumulated values of histogram^(1/t)
    for(int k=0; k<npoints; k++){
        PH[k]=H[nb*k];
        for(int n=1; n< nb; n++)
            if(nb*k+n < nbins) PH[k]+=H[nb*k+n];
        PH[k]=pow(PH[k], 1./t);
        sum+=PH[k];
    }
    for(int k=0; k< npoints; k++)
        S[k]=255.*PH[k]/(step*sum);
        
    //compute output
    double x0, x1, y0;        
    x0=L[0]; y0=0.;
    for(int k=0; k< npoints; k++){
        x1=L[k+1];
        
        for(int l=0; l< size; l++){            
            if(input[l] < Min) output[l]=0.;
            else if(input[l] >= Max) output[l]=255.;          
            else{
                if(input[l] >= x0 && input[l] < x1)
                    output[l]=y0+S[k]*(input[l]-x0);
            }
            
        }
        y0+=(x1-x0)*S[k];
        x0=x1;        
    }
    
    
    delete[] L;
    delete[] S;
    delete[] H;
    delete[] PH;
    delete[] sortdata;
    
}


//Estimate alpha parameter of f_alpha scaling function automatically 
//so median of intensity output = 127.5 
double estimate_parameter_alpha(double *input, int size, int nchannels,
                             double Min, double Max)
{
    double alpha;
    
    double *I;
    if (nchannels == 1) I=input;
    else {
        //compute intensity component
        I = new double[size];
        for (int n=0; n < size; n++) 
            I[n]=(input[n]+input[n+size]+input[n+2*size])/3.0;
    }

    double med=Median(I, size);
    alpha=MAX(log(2.)/(log(Max-Min)-log(med-Min)), 0.3);//clip the minimum value  
                                                    //of the exponent to 0.3
                                                    //to prevent excessive 
                                                    //enhancement of low 
                                                    //values

    if (nchannels > 1) delete[] I;
    
    return alpha;
}


//Implements f_alpha (Power) scaling function 
//(see Table 3 in IPOL accompanying paper)
//Parameters:
//alpha: factor in exponent of Equation 10 (see IPOL accompanying paper)
void f_alpha(double *output, double *input, double alpha, 
             int size, double Min, double Max)
{    
            
    if(Max <= Min) for(int l=0;l < size; l++) output[l]=input[l];
    else{
        double C=255./pow(Max-Min, alpha);
        for(int l=0; l< size;l++){
            if(input[l] < Min) output[l]=0.;
            else if(input[l] > Max) output[l]=255.;
            else output[l] = C*pow((input[l]-Min), alpha);
        }
    }
    
}


//Estimate A parameter of f_NR scaling function automatically 
//If median of input intensity takes a low value (median < (min+max)/2), then
//compute A such that the median is mapped to 127.5
//If median takes a high value (median >= (min+max)/2)), then
//use A=5 (almost linear scaling) to prevent excessive enhancement of high
//values
double estimate_parameter_NR(double *input, int size, int nchannels,
                             double Min, double Max)
{
    double A;
    
    double *I;
    if (nchannels == 1) I=input;
    else {
        //compute intensity component
        I = new double[size];
        for (int n=0; n < size; n++) 
            I[n]=(input[n]+input[n+size]+input[n+2*size])/3.0;
    }

    double med=Median(I, size);
    
    if (med < (Max+Min)/2) { 
        A=MAX((med-Min)/(Max+Min-2*med), 0.1); //clip the minimum value  
                                            //of the exponent to 0.1
                                            //to prevent excessive 
                                            //enhancement of low values
    } else A=5;
    

    if (nchannels > 1) delete[] I;
    
    return A;
}



//Implements f_NR (Naka-Rushton) scaling function 
//(see Table 3 in IPOL accompanying paper)
//Parameters:
//A: factor in Equation 11 (see IPOL accompanying paper)
void f_NR(double *output, double *input, double A, 
          int size, double Min, double Max)
{
        
    if(Max <= Min) for(int l=0;l < size; l++) output[l]=input[l];
    else {
        for(int l=0; l<size; l++){
            if(input[l] < Min) output[l]=0.;
            else if(input[l] > Max) output[l]=255.;
            else output[l]=255.*(A+1)*(input[l]-Min)/(A*(Max-Min)+input[l]-Min);
        }
    }
      
}

//Implements f_log (Logarithm) scaling function 
//(see Table 3 in IPOL accompanying paper)
void f_log(double *output, double *input, 
           int size, double Min, double Max)
{
    
    if(Max <= Min) for(int l=0;l < size; l++) output[l]=input[l];
    else {
        for(int l=0; l<size; l++){
            if(input[l]< Min) output[l]=0.;
            else if(input[l] > Max) output[l]=255.;
            else output[l]=255*log(input[l]-Min+1)/log(Max- Min+1);
        }
    }
   
}

//Implements f_L (Linear) scaling function 
//(see Table 3 in IPOL accompanying paper)
void f_L(double *output, double *input, 
         int size, double Min, double Max)
{	
	
	if(Max  <= Min)  for(int l=0; l < size; l++) output[l]=input[l];
	else{
		double scale=255./(Max-Min);
		for(int l=0; l< size; l++){
            if(input[l]< Min) output[l]=0.;
            else if(input[l] > Max) output[l]=255.;
            else output[l]=scale*(input[l]-Min);
		}
	}
	
}



